# Django
# Standard Library
from typing import Any, Dict, Iterable, Union

from django.http import HttpRequest, HttpResponse, StreamingHttpResponse
from django.template.response import TemplateResponse

# Local
from .renderers import Action, render_turbo_frame, render_turbo_stream


class TurboStreamResponseMixin:
    """Automatically sets the correct turbo-stream content type."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            content_type="text/html; turbo-stream; charset=utf-8", *args, **kwargs
        )


class TurboStreamStreamingResponse(TurboStreamResponseMixin, StreamingHttpResponse):
    """Handles turbo-stream responses. Generator should yield individual
    turbo-stream strings."""


class TurboStreamResponse(TurboStreamResponseMixin, HttpResponse):
    """Basic turbo-stream response."""

    def __init__(self, content: str = "", *, action: Action, target: str, **kwargs):
        super().__init__(
            render_turbo_stream(action, target, content), **kwargs,
        )


class TurboStreamTemplateResponse(TurboStreamResponseMixin, TemplateResponse):
    """Handles turbo-stream template response.

    Adds the following variables to the template:

    - **is_turbo_stream**
    - **turbo_stream_action**
    - **turbo_stream_target**

    """

    is_turbo_stream = True

    def __init__(
        self,
        request: HttpRequest,
        template: Union[str, Iterable[str]],
        context: Dict[str, Any],
        *,
        action: Action,
        target: str,
        **kwargs
    ):

        super().__init__(
            request,
            template,
            {
                **context,
                "turbo_stream_action": action.value,
                "turbo_stream_target": target,
                "is_turbo_stream": True,
            },
            **kwargs,
        )

        self._target = target
        self._action = action

    @property
    def rendered_content(self) -> str:
        return render_turbo_stream(
            action=self._action, target=self._target, content=super().rendered_content
        )


class TurboFrameResponse(HttpResponse):
    """Handles turbo-frame template response."""

    def __init__(self, content: str = "", *, dom_id: str, **kwargs):
        super().__init__(
            render_turbo_frame(dom_id, content), **kwargs,
        )


class TurboFrameTemplateResponse(TemplateResponse):
    """Handles turbo-stream template response.

    Adds the following variables to the template:

    - **is_turbo_frame**
    - **turbo_frame_dom_id**

    """

    is_turbo_frame = True

    def __init__(
        self,
        request: HttpRequest,
        template: Union[str, Iterable[str]],
        context: Dict[str, Any],
        *,
        dom_id,
        **kwargs
    ):

        super().__init__(
            request,
            template,
            {**context, "turbo_frame_dom_id": dom_id, "is_turbo_frame": True},
            **kwargs,
        )

        self._dom_id = dom_id

    @property
    def rendered_content(self) -> str:
        return render_turbo_frame(self._dom_id, super().rendered_content)



# Standard Library
import http
from typing import Union

# Django
from django.db.models import Model
from django.http import HttpResponseRedirect
from django.shortcuts import resolve_url


class HttpResponseSeeOther(HttpResponseRedirect):
    """Redirect with 303 status"""

    status_code = http.HTTPStatus.SEE_OTHER


def redirect_303(to: Union[str, Model], *args, **kwargs) -> HttpResponseSeeOther:
    """Sends an HTTP 303 redirect.

    :param to:  URL or view name or model instance. If model then calls `get_absolute_url()`.
    """
    return HttpResponseSeeOther(resolve_url(to, *args, **kwargs))



# Standard Library
from typing import Any, Dict

# Django
from django.http import HttpRequest, HttpResponse
from django.views.generic import (
    CreateView,
    DeleteView,
    FormView,
    TemplateView,
    UpdateView,
    View,
)

# Local
from . import Action
from .mixins import (
    TurboFormMixin,
    TurboFormModelMixin,
    TurboFrameResponseMixin,
    TurboFrameTemplateResponseMixin,
    TurboStreamResponseMixin,
    TurboStreamTemplateResponseMixin,
)


class TurboStreamView(TurboStreamResponseMixin, View):
    """Renders a simple turbo-stream view"""

    def dispatch(self, *args, **kwargs) -> HttpResponse:
        return self.render_turbo_stream_response()


class TurboStreamTemplateView(TurboStreamTemplateResponseMixin, TemplateView):
    """Renders response template inside <turbo-stream> tags. """

    def render_to_response(
        self, context: Dict[str, Any], **response_kwargs
    ) -> HttpResponse:
        return self.render_turbo_stream_template_response(context, **response_kwargs)


class TurboFormView(TurboFormMixin, FormView):
    ...


class TurboCreateView(TurboFormModelMixin, CreateView):
    ...


class TurboUpdateView(TurboFormModelMixin, UpdateView):
    ...


class TurboStreamDeleteView(TurboStreamResponseMixin, DeleteView):
    """Handles a delete action, returning an empty turbo-stream "remove"
    response. The target will be automatically resolved as {model_name}-{PK}.
    """

    turbo_stream_action = Action.REMOVE

    def get_turbo_stream_target(self) -> str:
        return f"{self.object._meta.model_name}-{self.object.pk}"

    def delete(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        # problem: object target needs to be defined *before* object is
        # deleted as ID will be None. So we need to get the response first
        # in order to resolve the target ID.
        self.object = self.get_object()
        response = self.render_turbo_stream_response()
        self.object.delete()
        return response


class TurboFrameView(TurboFrameResponseMixin, View):
    """Retuns a simple turbo-frame response."""

    def dispatch(self, *args, **kwargs) -> HttpResponse:
        return self.render_turbo_frame_response()


class TurboFrameTemplateView(TurboFrameTemplateResponseMixin, TemplateView):
    """Renders response template inside <turbo-frame> tags. """

    def render_to_response(
        self, context: Dict[str, Any], **response_kwargs
    ) -> HttpResponse:
        return self.render_turbo_frame_template_response(context, **response_kwargs)
